#include <stdbool.h>
#include <string.h>

#include "ltm/common.h"
#include "ltm/template.h"
#include "ltm/template_internal.h"

ltm_err ltm_template_init(ltm_template **out) {
  ltm_template *template = calloc(1, sizeof(ltm_template));

  if (template == NULL) {
    return ltm_err_failed_alloc;
  }

  *out = template;

  return ltm_err_ok;
}

bool ltm_template_next_placeholder(ltm_placeholder *ph, const char *s,
                                   size_t len) {
  ph->start = memchr(s, '{', len - 1);

  if ((ph->start != NULL) && (ph->start[1] == '{')) {
    size_t new_len = len - (ph->start - s);

    ph->end = memchr(ph->start + 2, '}', new_len - 3);

    // Non-terminated placeholders aren't valid
    if ((ph->end == NULL) || (ph->end[1] != '}')) {
      ph->type = ltm_placeholder_type_invalid;

      return true;
    }

    // End should point to final character
    ph->end++;

    // Parse the words
    ph->name.s = ph->start + 2;

    while ((*ph->name.s == ' ') && (ph->name.s != ph->end - 1)) {
      ph->name.s++;
    }

    // Placeholder is empty
    if (ph->name.s == ph->end - 1) {
      ph->type = ltm_placeholder_type_invalid;

      return true;
    }

    const char *ident = ph->name.s;

    while ((*ident != ' ') && (ident != ph->end - 1)) {
      ident++;
    }

    ph->name.len = ident - ph->name.s;

    while ((*ident == ' ') && (ident != ph->end - 1)) {
      ident++;
    }

    if (ident == ph->end - 1) {
      ph->type = ltm_placeholder_type_var;
    } else {
      // Further parse the identifier
      const char *temp = ident;

      while ((*temp != ' ') && (temp != ph->end - 1)) {
        temp++;
      }

      size_t ident_len = temp - ident;

      if (strncmp("loop", ident, ident_len) == 0) {
        ph->type = ltm_placeholder_type_loop_start;
      } else if (strncmp("end", ident, ident_len) == 0) {
        ph->type = ltm_placeholder_type_loop_end;
      } else {
        ph->type = ltm_placeholder_type_invalid;
      }
    }

    return true;
  }

  return false;
}

ltm_err ltm_template_compile_n(ltm_template **out, const char *s, size_t len) {
  ltm_template *template;
  LTM_RES(ltm_template_init(&template));

  ltm_placeholder ph;
  bool in_loop = false;
  const char *loop_start = NULL;
  size_t loop_depth = 0;
  size_t cur_loop_depth = 0;

  // TODO to ensure the loops are balanced, we should count how many loop starts
  // we have seen and only match a loop end if the number matches; this way, we
  // can allow arbitrarily nested loops

  while (ltm_template_next_placeholder(&ph, s, len)) {
    switch (ph.type) {
    case ltm_placeholder_type_invalid:
      return ltm_err_invalid_template;
    case ltm_placeholder_type_var:
      // TODO add var block
      break;
    case ltm_placeholder_type_loop_start:
      if (!in_loop) {
        loop_start = ph.end + 1;
        in_loop = true;
        loop_depth = cur_loop_depth;
      }
      cur_loop_depth++;
      break;
    case ltm_placeholder_type_loop_end:
      cur_loop_depth--;

      if (in_loop && (cur_loop_depth == loop_depth)) {
        size_t loop_len = ph.end - loop_start;
        // TODO recursive call to compile

        in_loop = false;
      }
      // We encountered a loop end without a start
      else {
        return ltm_err_invalid_template;
      }
      break;
    }

    len -= ph.end + 1 - ph.start;
    s = ph.end + 1;
  }

  // Unfinished loop
  if (in_loop) {
    return ltm_err_invalid_template;
  }

  return ltm_err_ok;
}

ltm_err ltm_template_compile(ltm_template **out, const char *template) {
  return ltm_template_compile_n(out, template, strlen(template));
}
